/*
 * Common data type definitions and functions
 *
 * Copyright IBM Corp. 2022
 *
 * s390-tools is free software; you can redistribute it and/or modify
 * it under the terms of the MIT license. See LICENSE for details.
 */
#ifndef PVATTEST_TYPES_H
#define PVATTEST_TYPES_H
#include <stdint.h>

/* Types to mark values as big endian. */
typedef uint16_t be16_t;
typedef uint32_t be32_t;
typedef uint64_t be64_t;

#endif /* PVATTEST_TYPES_H */
